
class SingleLookAtQueue(object):

	def __init__(self):
		self.__data = []
		self.__lookedAt = set()
	#

	def add(self, item):
		itemID = id(item)
		if itemID not in self.__lookedAt:
			self.__data.append(item)
			self.__lookedAt.add(itemID)
	#

	def addAll(self, items):
		for item in items:
			itemID = id(item)
			if itemID not in self.__lookedAt:
				self.__data.append(item)
				self.__lookedAt.add(itemID)
	#

	def retrieve(self):
		if self.__data:
			ret = self.__data[0]
			del self.__data[0]
			return ret
		else:
			return None
	#

	def isNotEmpty(self):
		return len(self.__data) > 0
	#

	def isEmpty(self):
		return len(self.__data) == 0
	#

#







from NodeMatrix import *









class Node(object):

	def __init__(self, name:str):
		self.name = name
	#

	def __str__(self):
		return self.name
	#

	def __repr__(self):
		return self.name
	#

#

r = Node("R")
a = Node("A")
b = Node("B")
c = Node("C")
d = Node("D")
e = Node("E")
f = Node("F")
g = Node("G")
h = Node("H")
i = Node("I")
j = Node("J")
k = Node("K")
l = Node("L")
m = Node("M")
n = Node("N")
o = Node("O")
p = Node("P")

nodes = [	r, a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p	]
nodeDict = {	node.name: node for node in nodes	}

links = [
	"AB",
	"AE",
	"AC",
	"OG",
	"EF",
	"CF",
	"FK",
	"FL",
	"FI",
	"IJ",
	"IK",
	"KM",
	"FH",
	"HN",
	"NL",
	"LH",	# cycle
	"LP",
	"AH",
]

matrix = NodeMatrix(nodes)
for link in links:
	matrix.set(nodeDict[link[0]], nodeDict[link[1]], True)

resultFile = matrix.convertTo("svg")
subprocess.Popen(["nohup", "/usr/bin/viewnior", resultFile], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

matrix.determineOrder(debugWriteFunction=print)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-



import time
import collections

import jk_logging
import jk_utils

from .TestContext import TestContext
from .Annotations import *
from .EnumProcessingState import EnumProcessingState
from .EnumEnabledState import EnumEnabledState






class ExceptionMatcher(object):

	def __init__(self, exceptionClass:typing.Union[str,type], exceptionData:dict):
		assert isinstance(exceptionClass, (str, type))
		self.exceptionClass = exceptionClass
		self.exceptionData = exceptionData if exceptionData else None
	#

	def matches(self, ee:Exception):
		if isinstance(self.exceptionClass, str):
			if ee.__class__.__name__ != self.exceptionClass:
				return False
		else:
			if ee.__class__ != self.exceptionClass:
				return False

		if self.exceptionData:
			for k, v in self.exceptionData.items():
				eev = getattr(ee, k, None)
				if eev != v:
					return False

		return True
	#

#



class TestCaseInstance(object):

	#
	# Constructor.
	#
	# @param	str testCaseName					The name of a test case.
	# @param	Annotation[] testCaseAspects		A (possibly empty) set of annotations that accompany the test case.
	# @param	callable testCaseCallable			The function to execute that implements the test.
	# @param	bool bEnabled						A flag that indicates if the test is enabled by default.
	#
	def __init__(self, bIsRoot:bool, testCaseName:str, testCaseAspects:collections.Sequence,
		testCaseCallable:collections.Callable, bEnabledByUser:bool):

		assert isinstance(bIsRoot, bool)

		if bIsRoot:
			# variables with test case definition or user provided data

			self.testCaseAspects = set()
			self.name = ""														# the name of the test
			self.testCaseCallable = None											# the function to execute to perform the test
			self.possibleExceptions = None
			self.variablesProvided = None
			self.description = None

			# variables filled at runtime

			self.enabledState = EnumEnabledState.ENABLED_IN_CONSEQUENCE
			self.processingState = EnumProcessingState.NOT_PROCESSED

		else:
			assert isinstance(testCaseName, str)
			assert isinstance(testCaseAspects, (list, tuple))
			assert callable(testCaseCallable)
			assert isinstance(bEnabledByUser, bool)

			# variables with test case definition or user provided data

			self.testCaseAspects = set(testCaseAspects) if testCaseAspects else set()			# stores all unprocessed dependencies
			self.name = testCaseName													# the name of the test
			self.testCaseCallable = testCaseCallable									# the function to execute to perform the test
			self.possibleExceptions = None
			self.variablesProvided = None
			self.description = None

			for a in self.testCaseAspects:

				if isinstance(a, ProvidesVariable):
					if self.variablesProvided is None:
						self.variablesProvided = []
					self.variablesProvided.append(a.varName)

				if isinstance(a, RaisesException):
					if self.possibleExceptions is None:
						self.possibleExceptions = []
					self.possibleExceptions.append(ExceptionMatcher(a.exceptionClass, a.exceptionData))

				if isinstance(a, Description):
					self.description = a.text

			#if self.possibleExceptions != None:
			#	for a in self.possibleExceptions:
			#		self.testCaseAspects.remove(a)

			# variables filled at runtime

			self.enabledState = EnumEnabledState.ENABLED_BY_USER if bEnabledByUser else EnumEnabledState.DISABLED
			self.processingState = EnumProcessingState.NOT_PROCESSED

		# variables filled at runtime

		self.dependsOn = set()		# stores the test cases
		self.requires = set()		# stores the test cases
		self.duration = None
		self.timeStamp = None
	#

	@property
	def isRoot(self):
		return self.testCaseCallable == None
	#

	def __str__(self):
		return self.name
	#

	def __repr__(self):
		return self.name
	#

	def runTest(self, data:dict, variables:dict, log) -> EnumProcessingState:
		nestedLog = log.descend("Running test: " + self.name + "()")

		bSuccess = False
		ctx = TestContext(data, variables, nestedLog)
		self.timeStamp = time.time()
		try:
			ret = self.testCaseCallable(True, ctx)
			self.duration = time.time() - self.timeStamp

			if self.possibleExceptions:
				# the test did not raise an exception!
				nestedLog.error("Test case " + repr(self.name) + " does not raise an exception!")
				self.processingState = EnumProcessingState.FAILED
				bSuccess = False

			else:

				if self.variablesProvided:
					# the test case should provide variables

					self.processingState = EnumProcessingState.SUCCEEDED
					bSuccess = True

					if isinstance(ret, dict):
						# the test case provides a dictionary with data.
						for varNameExpected in self.variablesProvided:
							if varNameExpected in ret:
								nestedLog.notice("Test case " + repr(self.name) + " provides variable " + repr(varNameExpected) + " as specified.")
								variables[varNameExpected] = ret[varNameExpected]
								del ret[varNameExpected]
							else:
								nestedLog.error("Test case " + repr(self.name) + " does not provide variable " + repr(varNameExpected) + " as specified!")
								self.processingState = EnumProcessingState.FAILED_CRITICALLY
								bSuccess = False
					else:
						# the test did not provide variables!
						nestedLog.error("Test case " + repr(self.name) + " does not provide variables!")
						self.processingState = EnumProcessingState.FAILED_CRITICALLY
						bSuccess = False

					if ret:
						nestedLog.warn("Test case " + repr(self.name) + " provides excessive variables: " + repr(list(ret.keys())))

				else:
					# the test case should not provide variables; if it does we ignore them;
					self.processingState = EnumProcessingState.SUCCEEDED
					bSuccess = True

		except Exception as ee:
			self.duration = time.time() - self.timeStamp

			bSuccess = False
			if self.possibleExceptions:
				for exceptionExpected in self.possibleExceptions:
					if exceptionExpected.matches(ee):
						bSuccess = True
						nestedLog.notice("Exception raised as expected: " + repr(ee))
						self.processingState = EnumProcessingState.SUCCEEDED
						break
			if not bSuccess:
				nestedLog.error(ee)
				self.processingState = EnumProcessingState.FAILED

		if bSuccess:
			nestedLog.success("Test passed.")
		else:
			nestedLog.error("TEST FAILED!")

		return self.processingState
	#

#














import jk_utils




class EnumProcessingState(jk_utils.EnumBase):

	FAILED_CRITICALLY = -2, "failed_critically"
	NOT_PROCESSED = -1, "not_processed"
	FAILED = 0, "failed"
	SUCCEEDED = 1, "succeeded"

#









import re

from .AssertionException import AssertionException







class _Assert(object):

	def __init__(self, log):
		if callable(log):
			self.__log = log
		else:
			self.__log = log
	#

	def isIn(self, value, valueList, message = None):
		Assert.l_isIn(self.__log, value, valueList, message)
	#

	def isNotIn(self, value, valueList, message = None):
		Assert.l_isNotIn(self.__log, value, valueList, message)
	#

	def raisesException(self, function, arguments, message = None):
		Assert.l_raisesException(self.__log, function, arguments, message)
	#

	def isCallable(self, value, message = None):
		Assert.l_isCallable(self.__log, value, message)
	#

	def isInstance(self, value, typeOrTypes, message = None):
		Assert.l_isInstance(self.__log, value, typeOrTypes, message)
	#

	def isRegExMatch(self, value, regexPattern, message = None):
		Assert.l_isRegExMatch(self.__log, value, regexPattern, message)
	#

	def isEqual(self, value, otherValue, message = None):
		Assert.l_isEqual(self.__log, value, otherValue, message)
	#

	def isGreater(self, value, otherValue, message = None):
		Assert.l_isGreater(self.__log, value, otherValue, message)
	#

	def isGreaterOrEqual(self, value, otherValue, message = None):
		Assert.l_isGreaterOrEqual(self.__log, value, otherValue, message)
	#

	def isSmaller(self, value, otherValue, message = None):
		Assert.l_isSmaller(self.__log, value, otherValue, message)
	#

	def isSmallerOrEqual(self, value, otherValue, message = None):
		Assert.l_isSmallerOrEqual(self.__log, value, otherValue, message)
	#

	def isNotEqual(self, value, otherValue, message = None):
		Assert.l_isNotEqual(self.__log, value, otherValue, message)
	#

	def isNone(self, value, message = None):
		Assert.l_isNone(self.__log, value, message)
	#

	def isNotNone(self, value, message = None):
		Assert.l_isNotNone(self.__log, value, message)
	#

	def isNotNoneOrEmpty(self, value, message = None):
		Assert.l_isNotNoneOrEmpty(self.__log, value, message)
	#

	def isTrue(self, value, message = None):
		Assert.l_isTrue(self.__log, value, message)
	#

	def isFalse(self, value, message = None):
		Assert.l_isFalse(self.__log, value, message)
	#

#










class Assert(object):

	@staticmethod
	def createCustomAssert(log):
		return _Assert(log)
	#

	"""
	@staticmethod
	def getAllBaseClasses(cls):
		# TODO: convert this to an iteration
		c = list(cls.__bases__)
		for base in c:
			c.extend(getAllBaseClasses(base))
		return c
	#
	"""

	@staticmethod
	def isIn(value, valueList, message = None, log = None, identifier:str = None):
		bSuccess = value in valueList

		if not bSuccess:
			if message is None:
				message = ""
			else:
				message += " :: "
			message = "ASSERTION ERROR :: " + message + "Value is " + repr(value) + " so value is not an element of list " + repr(valueList) + "!"
			if identifier:
				message = "<" + identifier + "> " + message
			if log != None:
				if callable(log):
					log(message)
				else:
					log.error(message)
			else:
				print(message)
			raise AssertionException(message)
	#

	@staticmethod
	def l_isIn(log, value, valueList, message = None, identifier:str = None):
		bSuccess = value in valueList

		if not bSuccess:
			if message is None:
				message = ""
			else:
				message += " :: "
			message = "ASSERTION ERROR :: " + message + "Value is " + repr(value) + " so value is not an element of list " + repr(valueList) + "!"
			if identifier:
				message = "<" + identifier + "> " + message
			if log != None:
				if callable(log):
					log(message)
				else:
					log.error(message)
			else:
				print(message)
			raise AssertionException(message)
	#

	@staticmethod
	def isNotIn(value, valueList, message = None, log = None, identifier:str = None):
		bSuccess = value not in valueList

		if not bSuccess:
			if message is None:
				message = ""
			else:
				message += " :: "
			message = "ASSERTION ERROR :: " + message + "Value is " + repr(value) + " so value is not an element of list " + repr(valueList) + "!"
			if identifier:
				message = "<" + identifier + "> " + message
			if log != None:
				if callable(log):
					log(message)
				else:
					log.error(message)
			else:
				print(message)
			raise AssertionException(message)
	#

	@staticmethod
	def l_isNotIn(log, value, valueList, message = None, identifier:str = None):
		bSuccess = value not in valueList

		if not bSuccess:
			if message is None:
				message = ""
			else:
				message += " :: "
			message = "ASSERTION ERROR :: " + message + "Value is " + repr(value) + " so value is not an element of list " + repr(valueList) + "!"
			if identifier:
				message = "<" + identifier + "> " + message
			if log != None:
				if callable(log):
					log(message)
				else:
					log.error(message)
			else:
				print(message)
			raise AssertionException(message)
	#

	@staticmethod
	def raisesException(function, arguments, message = None, log = None, identifier:str = None):
		bSuccess = True
		try:
			function(*arguments)
			bSuccess = False
		except Exception as ee:
			pass

		if not bSuccess:
			if message is None:
				message = ""
			else:
				message += " :: "
			message = "ASSERTION ERROR :: " + message + "No exception was raised!"
			if identifier:
				message = "<" + identifier + "> " + message
			if log != None:
				if callable(log):
					log(message)
				else:
					log.error(message)
			else:
				print(message)
			raise AssertionException(message)
	#

	@staticmethod
	def l_raisesException(log, function, arguments, message = None, identifier:str = None):
		bSuccess = True
		try:
			function(*arguments)
			bSuccess = False
		except Exception as ee:
			pass

		if not bSuccess:
			if message is None:
				message = ""
			else:
				message += " :: "
			message = "ASSERTION ERROR :: " + message + "No exception was raised!"
			if identifier:
				message = "<" + identifier + "> " + message
			if log != None:
				if callable(log):
					log(message)
				else:
					log.error(message)
			else:
				print(message)
			raise AssertionException(message)
	#

	@staticmethod
	def isCallable(value, message = None, log = None, identifier:str = None):
		if callable(value):
			return
		if message is None:
			message = ""
		else:
			message += " :: "
		message = "ASSERTION ERROR :: " + message + "Value is not a callable but of type " + str(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isCallable(log, value, message = None, identifier:str = None):
		if callable(value):
			return
		if message is None:
			message = ""
		else:
			message += " :: "
		message = "ASSERTION ERROR :: " + message + "Value is not a callable but of type " + str(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isInstance(value, typeOrTypes, message = None, log = None, identifier:str = None):
		if isinstance(value, typeOrTypes):
			return
		if issubclass(type(value), typeOrTypes):
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is of type " + str(type(value)) + " and not of type " + str(typeOrTypes)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isInstance(log, value, typeOrTypes, message = None, identifier:str = None):
		if isinstance(value, typeOrTypes):
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is of type " + str(type(value)) + " and not of type " + str(typeOrTypes)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isEqual(value, otherValue, message = None, log = None, identifier:str = None):
		if value == otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " and not " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isEqual(log, value, otherValue, message = None, identifier:str = None):
		if value == otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " and not " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isRegExMatch(value, regexPattern, message = None, log = None, identifier:str = None):
		m = re.match(regexPattern, value)
		if m:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which does not match " + repr(regexPattern) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isRegExMatch(log, value, regexPattern, message = None, identifier:str = None):
		m = re.match(regexPattern, value)
		if m:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which does not match " + repr(regexPattern) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isNotEqual(value, otherValue, message = None, log = None, identifier:str = None):
		if value != otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which is not expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isNotEqual(log, value, otherValue, message = None, identifier:str = None):
		if value != otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which is not expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isGreater(value, otherValue, message = None, log = None, identifier:str = None):
		if value > otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which not greater than " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isGreater(log, value, otherValue, message = None, identifier:str = None):
		if value > otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which not greater than " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isGreaterOrEqual(value, otherValue, message = None, log = None, identifier:str = None):
		if value >= otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which not greater or equal to " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isGreaterOrEqual(log, value, otherValue, message = None, identifier:str = None):
		if value >= otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which not greater or equal to " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isSmaller(value, otherValue, message = None, log = None, identifier:str = None):
		if value < otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which not smaller than " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isSmaller(log, value, otherValue, message = None, identifier:str = None):
		if value < otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which not smaller than " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isSmallerOrEqual(value, otherValue, message = None, log = None, identifier:str = None):
		if value <= otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which not smaller or equal to " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isSmallerOrEqual(log, value, otherValue, message = None, identifier:str = None):
		if value <= otherValue:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is " + repr(value) + " which not smaller or equal to " + repr(otherValue) + " as expected!"
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isNone(value, message = None, log = None, identifier:str = None):
		if value is None:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is not None as expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isNone(log, value, message = None, identifier:str = None):
		if value is None:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is not None as expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isNotNone(value, message = None, log = None, identifier:str = None):
		if value is not None:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is None which is not expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isNotNone(log, value, message = None, identifier:str = None):
		if value is not None:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is None which is not expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isNotNoneOrEmpty(value, message = None, log = None, identifier:str = None):
		if value:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is None or empty which is not expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isNotNoneOrEmpty(log, value, message = None, identifier:str = None):
		if value:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is None or empty which is not expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isTrue(value, message = None, log = None, identifier:str = None):
		if value is True:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is not true as expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isTrue(log, value, message = None, identifier:str = None):
		if value is True:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is not true as expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def isFalse(value, message = None, log = None, identifier:str = None):
		if value is False:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is not false as expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

	@staticmethod
	def l_isFalse(log, value, message = None, identifier:str = None):
		if value is False:
			return
		if message is None:
			message = "ASSERTION ERROR"
		else:
			message += " ::"
		message = "ASSERTION ERROR :: " + message + " Value is not false as expected: " + repr(value)
		if identifier:
			message = "<" + identifier + "> " + message
		if log != None:
			if callable(log):
				log(message)
			else:
				log.error(message)
		else:
			print(message)
		raise AssertionException(message)
	#

#











__version__ = "0.2021.5.25"



from .AssertionException import AssertionException
from .Assert import Assert
from .AssertX import AssertX
from .Annotations import *

from .TestDriver import TestDriver
from .TestResultCollection import TestResultCollection
from .TestResult import TestResult
from .TestContext import TestContext
from .EnumProcessingState import EnumProcessingState
from .EnumEnabledState import EnumEnabledState

from .TestReporterHTML import TestReporterHTML




import os
import sys
import subprocess

import copy
import graphviz

import jk_temporary




class GraphCycleError(Exception):

	def __init__(self, data):
		self.data = data
	#

#



class NodeMatrix(object):

	def __init__(self, nodeObjs = None):
		self.__nodeObjs = []
		self.__nodeObjIDsToIndex = {}
		self.__statusNodeProcessed = []
		self.__statusIncomingNodesToProcess = []
		self.__countIncomingNodes = []
		self.__countOutgoingNodes = []
		self.__matrix = []
		self.__count = 0

		if nodeObjs:
			self.addNodes(nodeObjs)
	#

	@property
	def size(self):
		return self.__count
	#

	def nodes(self):
		return tuple(self.__nodeObjs)
	#

	def addNode(self, nodeObj):
		assert nodeObj is not None

		self.__nodeObjIDsToIndex[id(nodeObj)] = self.__count
		self.__nodeObjs.append(nodeObj)
		self.__statusNodeProcessed.append(False)
		self.__statusIncomingNodesToProcess.append(0)
		self.__countIncomingNodes.append(0)
		self.__countOutgoingNodes.append(0)

		for row in self.__matrix:
			row.append(False)
		self.__count += 1
		newRow = [ False ] * self.__count
		self.__matrix.append(newRow)
	#

	def addNodes(self, nodeObjs):
		assert isinstance(nodeObjs, (tuple, list))

		countAdditional = len(nodeObjs)
		if countAdditional == 0:
			return

		for nodeObj in nodeObjs:
			assert nodeObj is not None

		i = 0
		for nodeObj in nodeObjs:
			self.__nodeObjIDsToIndex[id(nodeObj)] = self.__count + i
			i += 1
		self.__nodeObjs.extend(nodeObjs)
		self.__statusIncomingNodesToProcess.extend([ 0 ] * len(nodeObjs))
		self.__statusNodeProcessed.extend([ True ] * len(nodeObjs))
		self.__countIncomingNodes.extend([ 0 ] * len(nodeObjs))
		self.__countOutgoingNodes.extend([ 0 ] * len(nodeObjs))

		for row in self.__matrix:
			for j in range(0, countAdditional):
				row.append(False)
		self.__count += countAdditional
		for j in range(0, countAdditional):
			newRow = [ False ] * self.__count
			self.__matrix.append(newRow)
	#

	def set(self, fromNode, toNode, bHasEdge:bool):
		assert fromNode is not None
		assert toNode is not None
		assert isinstance(bHasEdge, bool)

		nFrom = self.__nodeObjIDsToIndex[id(fromNode)]
		#nFrom = self.__nodeObjs.index(fromNode)
		nTo = self.__nodeObjIDsToIndex[id(toNode)]
		#nTo = self.__nodeObjs.index(toNode)

		bAlreadyHasEdge = self.__matrix[nFrom][nTo]
		if bAlreadyHasEdge == bHasEdge:
			return

		self.__matrix[nFrom][nTo] = bHasEdge
		if bHasEdge:
			self.__countIncomingNodes[nTo] += 1
			self.__countOutgoingNodes[nFrom] += 1
		else:
			self.__countIncomingNodes[nTo] -= 1
			self.__countOutgoingNodes[nFrom] -= 1
	#

	def get(self, fromNode, toNode):
		assert fromNode is not None
		assert toNode is not None

		nFrom = self.__nodeObjIDsToIndex[id(fromNode)]
		#nFrom = self.__nodeObjs.index(fromNode)
		nTo = self.__nodeObjIDsToIndex[id(toNode)]
		#nTo = self.__nodeObjs.index(toNode)

		return self.__matrix[nFrom][nTo]
	#

	#
	# Analyses the graph. Detect cylces and determine the order of execution.
	#
	# If a cycle is detected a <c>GraphCycleError</c> is raised. The argument for the exception is the node that has been identified
	# as part of a cycle.
	#
	# @param		callable debugWriteFunction		A function to invoke to output debug messages.
	#												<c>None</c> is specified by default in order to deactivate this feature.
	#												For example you can use <c>print</c> as argument here to write to STDOUT.
	# @return		obj[]		Returns the node objects in the correct order (if no error occurred).
	#
	def determineOrder(self, debugWriteFunction=None):
		bNodeCompleted = [ False ] * self.__count
		nIncomingEdgesLeft = copy.deepcopy(self.__countIncomingNodes)
		nodeValues = [ None ] * self.__count
		indicesOfNodesNotYetProcessed = set(range(0, self.__count))

		listOfRootNodes = self.__getNodeIndicesWithoutIncomingEdges()
		if len(listOfRootNodes) == 0:
			raise Exception("No root node(s)!")

		currentGen = listOfRootNodes
		orderIndex = 1
		retDict = {}
		while len(currentGen) > 0:
			if debugWriteFunction:
				debugWriteFunction("Now processing this generation of nodes: " + str(self.__i2n(currentGen)))
			nextGen = set()
			for n in currentGen:
				currentNode = self.__nodeObjs[n]
				if debugWriteFunction:
					debugWriteFunction("\tLooking at node: " + str(currentNode) + ", index " + str(n))
				if bNodeCompleted[n]:
					if debugWriteFunction:
						debugWriteFunction("\t\tNode is already completed. Skipping")
					#raise GraphCycleError(currentNode)
				elif nIncomingEdgesLeft[n] > 0:
					if debugWriteFunction:
						debugWriteFunction("\t\tNode has more than one unprocessed incoming edges. Skipping.")
					pass
				else:
					# this is the last edge: process this node
					if debugWriteFunction:
						debugWriteFunction("\t\tNode has zero unprocessed incoming edges => Now procesing this node.")
					nodeValues[n] = orderIndex
					if debugWriteFunction:
						debugWriteFunction("\t\tMarking node as completed: " + str(currentNode))
					bNodeCompleted[n] = True
					outgoingNodeIndices = self.__getOutgoingIndices(currentNode)
					if debugWriteFunction:
						debugWriteFunction("\t\tNode has outgoing edges to nodes: " + str(self.__i2n(outgoingNodeIndices)))
					for i in outgoingNodeIndices:
						nextGen.add(i)
						if debugWriteFunction:
							debugWriteFunction("\t\tReducing unprocessed edge count of node " + str(self.__nodeObjs[i])
								+ " from " + str(nIncomingEdgesLeft[i]) + " to " + str(nIncomingEdgesLeft[i] - 1)) 
						nIncomingEdgesLeft[i] -= 1
					retDict[orderIndex] = currentNode
					orderIndex += 1
					indicesOfNodesNotYetProcessed.remove(n)
			currentGen = nextGen

		if indicesOfNodesNotYetProcessed:
			# there are unprocessed nodes. that means: we did not reach all nodes! => cycle
			for i in indicesOfNodesNotYetProcessed:
				if nIncomingEdgesLeft[i] > 0:
					raise GraphCycleError(self.__nodeObjs[i])

		return [ retDict[i] for i in range(1, self.__count + 1) ]
	#

	def __i2n(self, indices):
		return [ self.__nodeObjs[i] for i in indices ]
	#

	def getNodesWithoutIncomingEdges(self):
		ret = []

		for n in range(0, self.__count):
			bHasIncoming = False
			for i in range(0, self.__count):
				if self.__matrix[i][n]:
					bHasIncoming = True
					break
			if not bHasIncoming:
				ret.append(self.__nodeObjs[n])

		return ret
	#

	def __getNodeIndicesWithoutIncomingEdges(self):
		ret = []

		for n in range(0, self.__count):
			bHasIncoming = False
			for i in range(0, self.__count):
				if self.__matrix[i][n]:
					bHasIncoming = True
					break
			if not bHasIncoming:
				ret.append(n)

		return ret
	#

	def getNodesWithoutOutgoingEdges(self):
		ret = []

		for n in range(0, self.__count):
			row = self.__matrix[n]
			bHasOutgoing = False
			for i in range(0, self.__count):
				if row[i]:
					bHasOutgoing = True
					break
			if not bHasOutgoing:
				ret.append(self.__nodeObjs[n])

		return ret
	#

	def __getNodeIndicesWithoutOutgoingEdges(self):
		ret = []

		for n in range(0, self.__count):
			row = self.__matrix[n]
			bHasOutgoing = False
			for i in range(0, self.__count):
				if row[i]:
					bHasOutgoing = True
					break
			if not bHasOutgoing:
				ret.append(n)

		return ret
	#

	def getOutgoing(self, fromNode):
		assert fromNode is not None

		# n = self.__nodeObjs.index(fromNode)
		n = self.__nodeObjIDsToIndex[id(fromNode)]

		ret = []
		row = self.__matrix[n]
		for i in range(0, self.__count):
			if row[i]:
				ret.append(self.__nodeObjs[i])
		return ret
	#

	def __getOutgoingIndices(self, fromNode):
		assert fromNode is not None

		# n = self.__nodeObjs.index(fromNode)
		n = self.__nodeObjIDsToIndex[id(fromNode)]

		ret = []
		row = self.__matrix[n]
		for i in range(0, self.__count):
			if row[i]:
				ret.append(i)
		return ret
	#

	def getIncoming(self, toNode):
		assert toNode is not None

		# n = self.__nodeObjs.index(toNode)
		n = self.__nodeObjIDsToIndex[id(toNode)]

		ret = []
		for i in range(0, self.__count):
			if self.__matrix[i][n]:
				ret.append(self.__nodeObjs[i])
		return ret
	#

	def __getIncomingIndices(self, toNode):
		assert toNode is not None

		# n = self.__nodeObjs.index(toNode)
		n = self.__nodeObjIDsToIndex[id(toNode)]

		ret = []
		for i in range(0, self.__count):
			if self.__matrix[i][n]:
				ret.append(i)
		return ret
	#

	def convertTo(self, targetFormat:str, nodeColoringCallback = None):
		return _NODE_MATRIX_CONVERTER.convert(self, targetFormat, nodeColoringCallback=nodeColoringCallback)
	#

#



class _GraphVizConverter(object):

	@property
	def formats(self):
		return [ "png", "svg", "dot" ]
	#

	def convert(self, nodeMatrix:NodeMatrix, targetFormat:str, tempDirPath:str, nodeColoringCallback):
		assert isinstance(nodeMatrix, NodeMatrix)
		assert isinstance(targetFormat, str)
		assert isinstance(tempDirPath, str)

		ret = [
			"strict digraph {",
			"\trankdir=LR;"
			#"\trankdir=TB;"
		]

		ret.append("")

		allNodes = nodeMatrix.nodes()

		for n, node in enumerate(allNodes):
			s = "\t" + str(n) + " [ label=\"" + str(node) + "\""
			if nodeColoringCallback:
				bgColor, textColor, lineColor = nodeColoringCallback(node)
				if bgColor:
					s += " style=filled fillcolor=\"" + bgColor + "\""
				if textColor:
					s += " fontcolor=\"" + textColor + "\""
				if lineColor:
					s += " color=\"" + lineColor + "\""
			s += " ]"
			ret.append(s)

		ret.append("")

		for nFrom, fromNode in enumerate(allNodes):
			toNodes = nodeMatrix.getOutgoing(fromNode)
			for toNode in toNodes:
				nTo = allNodes.index(toNode)
				ret.append("\t" + str(nFrom) + " -> " + str(nTo))

		ret.append("}")
		graphVizSource = "\n".join(ret)
		tmpFilePath = jk_temporary.createTempFilePath(baseDirPath=tempDirPath, postfix=".dot")
		with open(tmpFilePath, "w") as f:
			f.write(graphVizSource)

		if targetFormat == "dot":
			return tmpFilePath
		else:
			resultFilePath = graphviz.render("dot", targetFormat, tmpFilePath)
			os.unlink(tmpFilePath)
			return resultFilePath
	#

#



class NodeMatrixConverterMgr(object):

	def __init__(self, tempDirPath:str = "/tmp"):
		if tempDirPath:
			assert isinstance(tempDirPath, str)
			assert os.path.isdir(tempDirPath)
			self.__tempDirPath = tempDirPath
		else:
			self.__tempDirPath = "/tmp"

		self.__formatMap = {}
	#

	def register(self, converter):
		for outputFormat in converter.formats:
			if outputFormat not in self.__formatMap:
				self.__formatMap[outputFormat] = converter
	#

	def convert(self, nodeMatrix:NodeMatrix, targetFormat:str, nodeColoringCallback = None):
		assert isinstance(nodeMatrix, NodeMatrix)
		assert isinstance(targetFormat, str)
		if nodeColoringCallback:
			assert callable(nodeColoringCallback)

		converter = self.__formatMap.get(targetFormat, None)
		if converter:
			return converter.convert(nodeMatrix, targetFormat, self.__tempDirPath, nodeColoringCallback)
		else:
			raise Exception("No conveter available for format: " + targetFormat)
	#

#



_NODE_MATRIX_CONVERTER = NodeMatrixConverterMgr()
_NODE_MATRIX_CONVERTER.register(_GraphVizConverter())












import jk_terminal_essentials

from .SimpleTextLineOutputBuffer import SimpleTextLineOutputBuffer






class OutputComparison(object):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self):
		self.buf1intro = SimpleTextLineOutputBuffer()
		self.buf1 = SimpleTextLineOutputBuffer()
		self.buf2intro = SimpleTextLineOutputBuffer()
		self.buf2 = SimpleTextLineOutputBuffer()
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def dump(self) -> bool:
		wbuf1 = max(self.buf1intro.width, self.buf1.width)
		wbuf2 = max(self.buf2intro.width, self.buf2.width)
		totalWidth = 4 + 3 + wbuf1 + 3 + wbuf2 + 2

		_colFrame = jk_terminal_essentials.FGCOLOR_LIGHT_CYAN

		print()
		print(_colFrame + "=" * totalWidth + jk_terminal_essentials.STYLE_RESET)

		for rowNo in range(max(self.buf1intro.height, self.buf2intro.height)):
			t1 = self.buf1intro.get(rowNo)
			t2 = self.buf2intro.get(rowNo)
			s = [
				"    ",

				_colFrame,
				" | ",
				jk_terminal_essentials.STYLE_RESET,

				t1,
				" " * (wbuf1 - len(t1)),

				_colFrame,
				" | ",
				jk_terminal_essentials.STYLE_RESET,

				t2,
				" " * (wbuf2 - len(t2)),

				_colFrame,
				" |",
				jk_terminal_essentials.STYLE_RESET,
			]
			print("".join(s))

		s = [
			_colFrame,
			"    ",
			" | ",
			"~" * wbuf1,
			" | ",
			"~" * wbuf2,
			" |",
			jk_terminal_essentials.STYLE_RESET,
		]
		print("".join(s))

		bHadErrors = False
		for rowNo in range(max(self.buf1.height, self.buf2.height)):
			t1 = self.buf1.get(rowNo)
			t2 = self.buf2.get(rowNo)
			bEq = t1 == t2
			if not bEq:
				bHadErrors = True

			s = [
				jk_terminal_essentials.FGCOLOR_LIGHT_YELLOW,
				str(rowNo).rjust(4, " "),

				_colFrame,
				" | ",

				jk_terminal_essentials.BGCOLOR_DARK_GRAY if bEq else jk_terminal_essentials.BGCOLOR_RED,
				jk_terminal_essentials.FGCOLOR_LIGHT_GREEN if bEq else jk_terminal_essentials.FGCOLOR_WHITE,
				t1,
				jk_terminal_essentials.STYLE_RESET,
				" " * (wbuf1 - len(t1)),

				_colFrame,
				" | ",

				jk_terminal_essentials.BGCOLOR_DARK_GRAY if bEq else jk_terminal_essentials.BGCOLOR_RED,
				jk_terminal_essentials.FGCOLOR_LIGHT_GREEN if bEq else jk_terminal_essentials.FGCOLOR_WHITE,
				t2,
				jk_terminal_essentials.STYLE_RESET,
				" " * (wbuf2 - len(t2)),

				_colFrame,
				" |",
				jk_terminal_essentials.STYLE_RESET,
			]
			print("".join(s))

		print(_colFrame + "=" * totalWidth + jk_terminal_essentials.STYLE_RESET)
		print()

		return bHadErrors
	#

	def compare(self) -> bool:
		for rowNo in range(max(self.buf1.height, self.buf2.height)):
			t1 = self.buf1.get(rowNo)
			t2 = self.buf2.get(rowNo)
			bEq = t1 == t2
			if not bEq:
				return True

		return False
	#

#



















#!/usr/bin/env python3
# -*- coding: utf-8 -*-



import time
import collections
import typing

import jk_logging

from .Annotations import *
from .TestCaseInstance import *
from .NodeMatrix import *
from .SingleLookAtQueue import *





class TestCaseCollection(object):

	def __init__(self,
		allTestCases:typing.Sequence,
		nodeMatrix:NodeMatrix,
		rootNode:TestCaseInstance,
		enabledTestCasesToRunInDefinedOrder:typing.Sequence):

		assert isinstance(allTestCases, (tuple, list, set))
		assert isinstance(nodeMatrix, NodeMatrix)
		assert isinstance(rootNode, TestCaseInstance)
		assert isinstance(enabledTestCasesToRunInDefinedOrder, (tuple, list))

		self.__allTestCases = allTestCases
		self.__nodeMatrix = nodeMatrix
		self.__rootNode = rootNode
		self.__enabledTestCasesToRunInDefinedOrder = enabledTestCasesToRunInDefinedOrder
	#

	@property
	def _nodeMatrix(self):
		return self.__nodeMatrix
	#

	@property
	def isFailed(self):
		return self.__enabledTestCasesToRunInDefinedOrder is None
	#

	@property
	def isReady(self):
		return self.__enabledTestCasesToRunInDefinedOrder is not None
	#

	@property
	def allTestCases(self):
		return self.__allTestCases
	#

	@property
	def enabledTestCasesToRunInDefinedOrder(self):
		return self.__enabledTestCasesToRunInDefinedOrder
	#

	#
	# Derives test case instances from the function.
	#
	@staticmethod
	def __parseTestCallable(theCallable:collections.Callable, bEnabled:bool) -> TestCaseInstance:
		if not callable(theCallable):
			raise Exception("Not a callable: " + repr(theCallable))
		assert isinstance(bEnabled, bool)

		bIsObject = True
		orgObj = None
		try:
			theCallable.__name__ == "TestCaseWrapper"
			bIsObject = False
		except:
			pass

		if bIsObject:
			orgObj = theCallable
			theCallable = theCallable.__call__

		if theCallable.__name__ == "TestCaseWrapper":
			testCaseAspects, testCaseName = theCallable(False)

			if bIsObject:
				newTestCaseName = None
				try:
					newTestCaseName = orgObj.name()
				except:
					pass
				if newTestCaseName is None:
					try:
						newTestCaseName = orgObj.name
					except:
						pass
				if newTestCaseName:
					testCaseName = newTestCaseName
				else:
					objName = theCallable.__class__.__name__
					if testCaseName == "__call__":
						testCaseName = objName
					else:
						testCaseName = objName + "." + testCaseName

			return TestCaseInstance(False, testCaseName, testCaseAspects, theCallable, bEnabled)
		else:
			# NOTE: we accept methods as well that are not a test case
			# TODO: is that reasonable?
			testCaseName = str(theCallable.__name__)
			return TestCaseInstance(False, testCaseName, None, theCallable, bEnabled)
	#

	#
	# This method creates a test case collection. While doing so various checks are performed, required test cases
	# will get enabled that might not have been enabled yet and cycles are detected. The order in which the tests
	# can be run will be determined.
	#
	# @param		list testsToRunTuples	A list of test functions to execute or a list of tuples with test functions and boolean enabled/disabled flag.
	# @return		TestCaseCollection		Returns a test case collection object. Please note that if cycles are detected
	#										a test case collection will be returned as well which then will be marked as
	#										being erroneous. Use <c>testCaseCollection.isReady</c> or
	#										<c>testCaseCollection.isFailed</c> to detect this.
	#
	@staticmethod
	def compile(testsToRunTuples:list, log = None):
		if len(testsToRunTuples) == 0:
			raise Exception("No tests specified!")

		if isinstance(testsToRunTuples, (tuple, list)):
			if callable(testsToRunTuples[0]):
				testsToRunTuples = [ (x, True) for x in testsToRunTuples ]

		# build all test case instances

		if log:
			log.info("Creating test case instances ...")

		allTestCases = []
		allTestCasesByName = {}
		allTestCasesByProvidedVariables = {}
		for testCallable, bEnabled in testsToRunTuples:
			testCaseInstance = TestCaseCollection.__parseTestCallable(testCallable, bEnabled)
			if testCaseInstance.name in allTestCasesByName:
				raise Exception("Duplicate test cases: " + testCaseInstance.name)
			allTestCasesByName[testCaseInstance.name] = testCaseInstance
			allTestCases.append(testCaseInstance)
			for aspect in testCaseInstance.testCaseAspects:
				if isinstance(aspect, ProvidesVariable):
					if aspect.varName in allTestCasesByProvidedVariables:
						raise Exception("Two test cases provide the same variable " + repr(aspect.varName) + ": "
							+ repr(testCaseInstance.name) + " and " + repr(allTestCasesByProvidedVariables[aspect.varName].name))
					allTestCasesByProvidedVariables[aspect.varName] = testCaseInstance

		# add all test cases to the graph

		log2 = log.descend("Building graph of test cases ...") if log else None

		rootNode = TestCaseInstance(True, None, None, None, None)
		nodeMatrix = NodeMatrix()
		nodeMatrix.addNode(rootNode)
		nodeMatrix.addNodes(allTestCases)
		for i in range(0, len(allTestCases)):
			nodeMatrix.set(rootNode, allTestCases[i], True)

		# add all dependencies to the graph

		testCasesToActivate = SingleLookAtQueue()
		for a in allTestCases:
			for aspect in a.testCaseAspects:

				# A.runBefore(B) =>		A -> B
				if isinstance(aspect, RunBefore):
					b = allTestCasesByName.get(aspect.testName, None)
					if b is None:
						if log2:
							log2.error("Test case '" + testCaseInstance.name + "' refers to a non-existing test case named '" + aspect.testName + "'!")
						return False
					if log2:
						log2.notice("Constraint: Test case " + repr(a.name) + " should run before test case " + repr(b.name) + ".")
					nodeMatrix.set(a, b, True)

				# A.runAfter(B) =>		B -> A
				elif isinstance(aspect, RunAfter):
					b = allTestCasesByName.get(aspect.testName, None)
					if b is None:
						if log2:
							log2.error("Test case '" + testCaseInstance.name + "' refers to a non-existing test case named '" + aspect.testName + "'!")
						return False
					if log2:
						log2.notice("Constraint: Test case " + repr(a.name) + " should run after test case " + repr(b.name) + ".")
					nodeMatrix.set(b, a, True)

				# A.requires(B) =>		B -> A
				elif isinstance(aspect, Requires):
					b = allTestCasesByName.get(aspect.testName, None)
					if b is None:
						if log2:
							log2.error("Test case '" + testCaseInstance.name + "' refers to a non-existing test case named '" + aspect.testName + "'!")
						return False
					if log2:
						log2.notice("Constraint: Test case " + repr(a.name) + " requires success of prior test case " + repr(b.name) + ".")
					nodeMatrix.set(b, a, True)
					a.dependsOn.add(b)
					a.requires.add(b)
					testCasesToActivate.add(b)

				elif isinstance(aspect, ProvidesVariable):
					pass

				# A.requiresVariable(varName) =>		B -> A
				elif isinstance(aspect, RequiresVariable):
					b = allTestCasesByProvidedVariables.get(aspect.varName, None)
					if b is None:
						if log2:
							log2.error("Test case '" + testCaseInstance.name + "' refers to variable named '" + a.varName + "' which no test case provides!")
						return False
					if log2:
						log2.notice("Constraint: Test case " + repr(a.name) + " requires variable created by test case " + repr(b.name) + ".")
					nodeMatrix.set(b, a, True)
					a.dependsOn.add(b)
					a.requires.add(b)
					testCasesToActivate.add(b)

				# ignore
				elif isinstance(aspect, RaisesException):
					pass

				# ignore
				elif isinstance(aspect, Description):
					pass

				else:
					raise Exception()

		# clean excessive root node connections

		if log:
			log.notice("Removing excessive nodes previously introduced for building the graph ...")

		TestCaseCollection.__cleanExcessiveRootNodeConnections(nodeMatrix)

		# determine the order of processing; we need to do this now as further processing requires to definitively have no cycles

		log2 = log.descend("Calculating order of tests ...") if log else None

		try:
			allTestCases = nodeMatrix.determineOrder()
		except GraphCycleError as ee:
			if log2:
				log2.error("Cycle detected at test case: " + ee.data.name)
			return TestCaseCollection(allTestCases, nodeMatrix, rootNode, None)

		#log.notice("Total number of test cases: " + str(len(allTestCases)))

		# activate test cases scheduled for activation; recursively enable test cases as necessary

		log2 = log.descend("Activating disabled test cases that are required ...") if log else None

		while testCasesToActivate.isNotEmpty():
			testCaseInstance = testCasesToActivate.retrieve()
			if testCaseInstance.enabledState == EnumEnabledState.DISABLED:
				testCaseInstance.enabledState = EnumEnabledState.ENABLED_IN_CONSEQUENCE
				testCasesToActivate.addAll(testCaseInstance.requires)
				if log2:
					log2.notice("Enabling: " + testCaseInstance.name)

		# collect all enabled test cases

		if log:
			log.info("Preparing everything to return the results ...")

		enabledTestCasesToRunInDefinedOrder = [
			testCaseInstance for testCaseInstance in allTestCases
				if testCaseInstance.enabledState in [ EnumEnabledState.ENABLED_BY_USER, EnumEnabledState.ENABLED_IN_CONSEQUENCE ]
		]

		#rootNode.enabledState = EnumEnabledState.DISABLED
		#enabledTestCasesToRunInDefinedOrder.remove(rootNode)

		# return TestCaseCollection-object

		return TestCaseCollection(allTestCases, nodeMatrix, rootNode, enabledTestCasesToRunInDefinedOrder)
	#

	@staticmethod
	def __cleanExcessiveRootNodeConnections(nodeMatrix:NodeMatrix):
		allNodes = nodeMatrix.nodes()
		rootNode = allNodes[0]
		allNodesExceptRoot = nodeMatrix.nodes()[1:]

		for node in allNodesExceptRoot:
			allIncomingNodes = nodeMatrix.getIncoming(node)
			allIncomingNodes.remove(rootNode)
			if len(allIncomingNodes) > 0:
				nodeMatrix.set(rootNode, node, False)
	#

#





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import functools
import typing








class RequiresVariable(object):

	def __init__(self, varName):
		if isinstance(varName, str):
			self.varName = varName
		else:
			raise Exception("Variable name expected!")
	#

#



class ProvidesVariable(object):

	def __init__(self, varName):
		if isinstance(varName, str):
			self.varName = varName
		else:
			raise Exception("Variable name expected!")
	#

#



class RunAfter(object):

	def __init__(self, testName):
		if isinstance(testName, str):
			self.testName = testName
		else:
			raise Exception("Name of a test expected!")
	#

#



class Requires(object):

	def __init__(self, testName):
		if isinstance(testName, str):
			self.testName = testName
		else:
			raise Exception("Name of a test expected!")
	#

#



class RunBefore(object):

	def __init__(self, testName):
		if isinstance(testName, str):
			self.testName = testName
		else:
			raise Exception("Name of a test expected!")
	#

#



class RaisesException(object):

	def __init__(self, exceptionClass:typing.Union[str,type], **kwargs):
		if isinstance(exceptionClass, (str, type)):
			self.exceptionClass = exceptionClass
			self.exceptionData = kwargs if kwargs else None
		else:
			raise Exception("Exception class expected!")
	#

#



class Description(object):

	def __init__(self, text):
		if isinstance(text, str):
			if len(text) > 0:
				self.text = text
			else:
				raise Exception("Description is empty!")
		else:
			raise Exception("Description text expected!")
	#

#











def _appendTo(someItemOrItems, classToWrap, someList):
	if isinstance(someList, tuple):
		someList = list(someList)
	if isinstance(someList, list):
		if isinstance(someItemOrItems, (list, tuple)):
			for x in someItemOrItems:
				someList.append(classToWrap(someItemOrItems))
		else:
			someList.append(classToWrap(someItemOrItems))
		return someList
	else:
		raise ValueError()
#



class TestCase(object):

	__VALID_CLASSES = (ProvidesVariable, RequiresVariable, RunAfter, RunBefore, Requires, RaisesException, Description)

	def __init__(self,
		*testAspects,
		**kwargs
		):

		# check if test aspects are valid
		for a in testAspects:
			if not isinstance(a, TestCase.__VALID_CLASSES):
				raise Exception("Unknown test aspect: " + str(a))

		if "requires" in kwargs:
			testAspects = _appendTo(kwargs["requires"], Requires, testAspects)
		if "runBefore" in kwargs:
			testAspects = _appendTo(kwargs["runBefore"], RunBefore, testAspects)
		if "runAfter" in kwargs:
			testAspects = _appendTo(kwargs["runAfter"], RunAfter, testAspects)
		if "requiresVariable" in kwargs:
			testAspects = _appendTo(kwargs["requiresVariable"], RequiresVariable, testAspects)
		if "providesVariable" in kwargs:
			testAspects = _appendTo(kwargs["providesVariable"], ProvidesVariable, testAspects)
		if "description" in kwargs:
			testAspects = _appendTo(kwargs["description"], Description, testAspects)

		# remember them for later
		self.testAspects = testAspects
	#

	def __get__(self, obj, objtype):
		return functools.partial(self.__call__, obj)
	#

	def __call__(self, original_func):
		decorator_self = self
		def TestCaseWrapper(*args, **kwargs):
			if isinstance(args[0], bool):
				# function
				assert len(args) >= 1
				args = list(args)
				if args[0]:
					del args[0]
					return original_func(*args,**kwargs)
				else:
					return (decorator_self.testAspects, original_func.__name__)
			else:
				# method
				assert len(args) >= 2
				assert isinstance(args[1], bool)
				args = list(args)
				if args[1]:
					del args[1]
					return original_func(*args,**kwargs)
				else:
					return (decorator_self.testAspects, original_func.__name__)
		return TestCaseWrapper
	#

#









import os
import sys
import subprocess

from .TestCaseInstance import *
from .NodeMatrix import NodeMatrix




class TestCollectionVisualizer(object):

	IMAGE_VIEWER_CANDIDATES = [
		"/usr/bin/viewnior",		# viewnior: https://siyanpanayotov.com/project/viewnior
		"/usr/bin/nomacs",			# nomacs: https://nomacs.org/
		"/usr/bin/geeqie",			# geeqie: http://geeqie.org/
		"/usr/bin/mirageiv",		# mirageiv: http://mirageiv.sourceforge.net/
		"/usr/bin/eom",				# eye of mate
		"/usr/bin/eog",				# eye of gnome
		"/usr/bin/feh",				# feh: https://feh.finalrewind.org/
		"/usr/bin/display",			# fallback
	]

	def __init__(self):
		pass
	#

	#
	# Determine the color a node should have during visualization
	#
	# @return	str bgColor			The background color selected (or <c>None</c>)
	# @return	str textColor		The text color selected (or <c>None</c>)
	# @return	str lineColor		The foreground color selected (or <c>None</c>)
	#
	def __calcNodeColorCallback(self, testCaseInstance:TestCaseInstance):
		if testCaseInstance.isRoot:
			# test is root
				# bgColor: light blue
				# textColor: dark blue
				# lineColor: dark blue
			return "#f0f0ff", "#f0f0ff", "#d0d0f0"

		if testCaseInstance.processingState == EnumProcessingState.NOT_PROCESSED:
			if testCaseInstance.enabledState == EnumEnabledState.ENABLED_BY_USER:
				# test enabled by user
					# bgColor: light green
					# textColor: dark green
					# lineColor: dark green
				return "#e0ffe0", "#40a040", "#40a040"
			elif testCaseInstance.enabledState == EnumEnabledState.ENABLED_IN_CONSEQUENCE:
				# test enabled in consequence
					# bgColor: light grayish green
					# textColor: dark grayish green
					# lineColor: dark grayish green
				return "#e0f0e0", "#80a080", "#80a080"
			elif testCaseInstance.enabledState == EnumEnabledState.DISABLED:
				# test disabled
					# bgColor: light gray
					# textColor: gray
					# lineColor: gray
				return "#e0e0e0", "#909090", "#909090"
			else:
				raise Exception()
		elif testCaseInstance.processingState == EnumProcessingState.SUCCEEDED:
			# succeeded
				# bgColor: dark green
				# textColor: white
				# lineColor: white
			return "#008000", "#ffffff", "#ffffff"
		elif testCaseInstance.processingState == EnumProcessingState.FAILED:
			# test failed
				# bgColor: dark red
				# textColor: white
				# lineColor: white
			return "#800000", "#ffffff", "#ffffff"
		elif testCaseInstance.processingState == EnumProcessingState.FAILED_CRITICALLY:
			# test failed
				# bgColor: dark red
				# textColor: white
				# lineColor: white
			return "#800000", "#ffffff", "#ffffff"
		else:
			raise Exception() 
	#

	def createSVG(self, collection):
		return collection._nodeMatrix.convertTo("svg", nodeColoringCallback=self.__calcNodeColorCallback)
	#

	def visualize(self, collection, imageViewerPath:str = None):
		if imageViewerPath is None:
			for filePath in TestCollectionVisualizer.IMAGE_VIEWER_CANDIDATES:
				if os.path.isfile(filePath):
					imageViewerPath = filePath
					break
		else:
			assert isinstance(imageViewerPath, str)

		if imageViewerPath is None:
			raise Exception("Autodetection of suitable image viewer failed!")
		elif os.path.isfile(imageViewerPath):
			resultFilePath = collection._nodeMatrix.convertTo("svg", nodeColoringCallback=self.__calcNodeColorCallback)
			subprocess.Popen(["nohup", imageViewerPath, resultFilePath], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
		else:
			raise Exception("No such image viewer: " + str(imageViewerPath))
	#

#





#!/usr/bin/env python3
# -*- coding: utf-8 -*-



import time
import collections
import subprocess
import queue
import datetime

import jk_logging

from .Annotations import *
from .TestCaseInstance import *
from .TestCaseCollection import *
from .NodeMatrix import *
from .TestResult import *
from .TestResultCollection import *




getCurrentTimeMillis = lambda: int(round(time.time() * 1000))






class TestDriver(object):

	def __init__(self):
		self.log = jk_logging.ConsoleLogger.create(logMsgFormatter = jk_logging.COLOR_LOG_MESSAGE_FORMATTER)
		self.__data = {}
	#

	@property
	def data(self):
		return self.__data
	#

	#
	# Run tests.
	#
	# @param	tuple<callable,bool>[] testsToRun		Tuples that define which tests to run. Each tuple contains the following entries:
	#													* The test case represented by a function to run.
	#													* A flag indicating if this test case is to be run.
	# @return	tuple<int,int,int,int,TestResult[]>		Returns a tuple of the following structure:
	#													* The number of tests performed (or even attempted to perform).
	#													* The number of tests skipped.
	#													* The number of tests succeeded.
	#													* The number of tests failed.
	#													* list of test result records.
	#
	def runTests(self, testsToRun) -> TestResultCollection:
		log2 = self.log.descend("Compiling test cases")
		testCaseCollection = TestCaseCollection.compile(testsToRun, log2)
		results = TestResultCollection(testCaseCollection, TestCollectionVisualizer())
		if testCaseCollection.isFailed:
			# prepairing tests failed!
			log2.error("Preparations failed!")
			return results
		log2.success("Preparations succeeded.")

		log2 = self.log.descend("Running tests ...")

		variables = {}
		theTestsToRun = testCaseCollection.enabledTestCasesToRunInDefinedOrder
		t0 = datetime.datetime.now()

		for testCaseInstance in theTestsToRun:
			if testCaseInstance.isRoot:
				continue

			logBuffer = jk_logging.BufferLogger.create()
			processingState = testCaseInstance.runTest(self.__data, variables, logBuffer)
			testResult = TestResult(testCaseInstance, logBuffer)
			logBuffer.forwardTo(log2)

			results.testResults.append(testResult)
			if processingState == EnumProcessingState.FAILED:
				results.countTestsFailed += 1
			elif processingState == EnumProcessingState.FAILED_CRITICALLY:
				results.countTestsFailed += 1
				break
			elif processingState == EnumProcessingState.SUCCEEDED:
				results.countTestsSucceeded += 1
			else:
				raise Exception()
			results.totalTestDuration += testResult.duration

		results.totalTestRuntime = (datetime.datetime.now() - t0).total_seconds()

		log2 = self.log.descend("Summary")
		log2.info("Number of tests performed: " + str(results.countTestsPerformed))
		if results.countTestsFailed > 0:
			log2.info("Number of tests succeeded: " + str(results.countTestsSucceeded))
			log2.error("Number of tests failed: " + str(results.countTestsFailed))
		else:
			log2.success("Number of tests succeeded: " + str(results.countTestsSucceeded))
			log2.info("Number of tests failed: " + str(results.countTestsFailed))
		log2.info("Total duration: " + jk_utils.formatTime(results.totalTestRuntime))
		log2.info("Average duration of single test: " + str(round(results.totalTestRuntime * 1000 / results.countTestsPerformed, 1)) + " ms")

		return results
	#

	"""
	def __runTests(self, testsToRun):
		t0 = getCurrentTimeMillis()

		self.log.info("Building list of all tests ...")
		testCaseCollection = TestCaseCollection()
		for testCallable, bEnabled in testsToRun:
			testCaseInstance = self.__parseTestCallable(testCallable, bEnabled)
			testCaseCollection.add(testCaseInstance)

		if not testCaseCollection.prepare(self.log.descend("Preparing test execution ...")):
			self.log.error("Failed to prepare all test cases for execution.")
			return False

		#for testCaseInstance, enabled in testCaseCollection.getTestTasksToProcess():
		#	print("-- " + testCaseInstance.name + " : " + str(enabled))

		t1 = getCurrentTimeMillis()

		testsPerformedList = []
		variables = {}
		countPerformed = 0
		countSkipped = 0
		countSucceeded = 0
		countFailed = 0
		for testCaseInstance, bEnabled in testCaseCollection.getTestTasksToProcess():
			if bEnabled:
				countPerformed += 1
				try:
					blog = jk_logging.BufferLogger.create()
					testCaseInstance.runTest(variables, jk_logging.MulticastLogger.create(self.log, blog))
					testsPerformedList.append(TestResult(testCaseInstance.name, testCaseInstance, blog, True))
					countSucceeded += 1
				except Exception as e:
					if e.__class__.__name__ == "_ExceptionInChildContextException":
						testsPerformedList.append(TestResult(testCaseInstance.name, testCaseInstance, blog, False))
						countFailed += 1
					else:
						raise
			else:
				countSkipped += 1

		t2 = getCurrentTimeMillis()

		# ----

		print()
		print()
		myLog = self.log.descend("Summary")
		myLog.info(str(t1 - t0) + " ms used to prepare and schedule the tests.")
		myLog.info(str(t2 - t1) + " ms used to run the tests.")
		myLog.info(str(countPerformed) + " tests attempted (= active tests).")
		myLog.info(str(countSucceeded) + " tests succeeded.")
		myLog.info(str(countFailed) + " tests failed.")
		myLog.info(str(countSkipped) + " tests skipped (= inactive tests).")

		return (countPerformed, countSkipped, countSucceeded, countFailed, testsPerformedList)
	#
	"""

#










from .TestResult import TestResult
from .TestCollectionVisualizer import TestCollectionVisualizer
from .TestCaseCollection import TestCaseCollection




class TestResultCollection(object):

	def __init__(self, testCaseCollection:TestCaseCollection, testCollectionVisualizer:TestCollectionVisualizer):
		self.__testCaseCollection = testCaseCollection
		self.__nodeMatrix = testCaseCollection._nodeMatrix
		self.testResults = []
		self.countTestsNotYetPerformed = 0
		self.countTestsSucceeded = 0
		self.countTestsFailed = 0
		self.totalTestDuration = 0
		self.totalTestRuntime = 0
		self.__testCollectionVisualizer = testCollectionVisualizer
	#

	@property
	def countTestsPerformed(self):
		return self.countTestsFailed + self.countTestsSucceeded
	#

	@property
	def _nodeMatrix(self):
		return self.__nodeMatrix
	#

	def createSVG(self):
		return self.__testCollectionVisualizer.createSVG(self)
	#

	def visualize(self, imageViewerPath:str = None):
		return self.__testCollectionVisualizer.visualize(self, imageViewerPath)
	#

#





################################################################################
################################################################################
###
###  This file is automatically generated. Do not change this file! Changes
###  will get overwritten! Change the source file for "setup.py" instead.
###  This is either 'packageinfo.json' or 'packageinfo.jsonc'
###
################################################################################
################################################################################


from setuptools import setup

def readme():
	with open("README.md", "r", encoding="UTF-8-sig") as f:
		return f.read()

setup(
	author = "Jürgen Knauth",
	author_email = "pubsrc@binary-overflow.de",
	classifiers = [
		"Development Status :: 5 - Production/Stable",
		"License :: OSI Approved :: Apache Software License",
		"Programming Language :: Python :: 3",
		"Topic :: Software Development :: Testing",
	],
	description = "This python module provides a simple to use infrastructure for running unit tests.",
	include_package_data = True,
	install_requires = [
		"jk_logging",
		"jk_temporary",
		"jk_json",
		"graphviz",
		"jinja2",
		"jk_terminal_essentials",
	],
	keywords = [
		"testing",
		"unittests",
		"tests",
	],
	license = "Apache2",
	name = "jk_testing",
	package_data = {
		"": [
			"data/*",
			"data/html_default/*",
			"data/html_default/files/*",
			"data/html_default/files/images/*",
			"data/html_default/templates/*",
		],
	},
	packages = [
		"jk_testing",
		"jk_testing.utils",
	],
	version = "0.2021.5.25",
	zip_safe = False,
	long_description = readme(),
	long_description_content_type="text/markdown",
)



import os
import sys
import shutil
import http.server
import socketserver
import webbrowser
import urllib
import posixpath

import jk_logging
import jk_json

from jinja2 import Environment, FileSystemLoader, select_autoescape, Template

from .TestResultCollection import *
from .TestResult import *




class ResultsHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):

	def __init__(self, rootDirPath:str, request, client_address, server):
		self.__rootDirPath = os.path.realpath(rootDirPath)
		super().__init__(request, client_address, server)
	#

	def translate_path(self, path):
		"""Translate a /-separated PATH to the local filename syntax.

		Components that mean special things to the local file system
		(e.g. drive or directory names) are ignored.  (XXX They should
		probably be diagnosed.)

		"""
		# abandon query parameters
		path = path.split('?',1)[0]
		path = path.split('#',1)[0]
		# Don't forget explicit trailing slash when normalizing. Issue17324
		trailing_slash = path.rstrip().endswith('/')
		try:
			path = urllib.parse.unquote(path, errors='surrogatepass')
		except UnicodeDecodeError:
			path = urllib.parse.unquote(path)
		path = posixpath.normpath(path)
		words = path.split('/')
		words = filter(None, words)
		path = self.__rootDirPath
		for word in words:
			if os.path.dirname(word) or word in (os.curdir, os.pardir):
				# Ignore components that are not a simple file/directory name
				continue
			path = os.path.join(path, word)
		if trailing_slash:
			path += '/'
		return path
	#	

#


class ResultsHTTPServer(socketserver.TCPServer):

	allow_reuse_address = True

	def __init__(self, rootDirPath:str):
		super().__init__(("", 9096), ResultsHTTPRequestHandler)
		self.__rootDirPath = rootDirPath
	#

	def finish_request(self, request, client_address):
		"""Finish one request by instantiating RequestHandlerClass."""
		self.RequestHandlerClass(self.__rootDirPath, request, client_address, self)
	#

#



class TestReporterHTML(object):

	def __init__(self):
		#self.__staticFilesDir = os.path.join(os.getcwd(), "files")
		self.__staticFilesDir = os.path.join(os.path.dirname(__file__), "data", "html_default", "files")

		#self.__templateDir = os.path.join(os.getcwd(), "templates")
		self.__templateDir = os.path.join(os.path.dirname(__file__), "data", "html_default", "templates")

		self.__env = Environment(
			loader=FileSystemLoader(self.__templateDir),
			autoescape=select_autoescape(["html", "xml"])
		)
		self.__templateTestCase = self.__env.get_template("testcase.html")
		self.__templateOverview = self.__env.get_template("index.html")
	#

	def report(self, testResultCollection:TestResultCollection, outDirPath:str="results", showInWebBrowser:bool=True):
		if not os.path.isabs(outDirPath):
			outDirPath = os.path.abspath(outDirPath)

		if not os.path.isdir(outDirPath):
			os.makedirs(outDirPath)
		shutil.rmtree(outDirPath)

		# ----

		shutil.copytree(self.__staticFilesDir, outDirPath)

		# ----

		for testResult in testResultCollection.testResults:
			self.__writeResultFile(testResult, outDirPath)

		self.__writeOverviewFile(testResultCollection, outDirPath)

		# ----

		if showInWebBrowser:
			httpd = ResultsHTTPServer(outDirPath)
			print("Running web server. For results see: http://localhost:9096/")
			webbrowser.open("http://localhost:9096/", new=1)
			httpd.serve_forever()
	#

	def __writeResultFile(self, testResult:TestResult, outDirPath:str):

		jsonTestRecord = {
				"id": "test_" + testResult.name,
				"file": "test_" + testResult.name + ".html",
				"name": testResult.name,
				"timeStamp": testResult.timeStamp,
				"enabledState": str(testResult.enabledState),
				"processingState": str(testResult.processingState),
				"duration": testResult.duration,
				"logBuffer": testResult.logBuffer.getDataAsPrettyJSON(),
				"description": testResult.description,
			}

		content = self.__templateTestCase.render(testRecord=jsonTestRecord)
		filePath = os.path.join(outDirPath, "test_" + testResult.name + ".html")
		print("Writing to: " + filePath)
		with open(filePath, "w") as f:
			f.write(content)
	#

	def __writeOverviewFile(self, testResultCollection:TestResultCollection, outDirPath:str):
		jsonTestRecords = []
		for testResult in testResultCollection.testResults:
			jsonTestRecords.append({
				"id": "test_" + testResult.name,
				"file": "test_" + testResult.name + ".html",
				"name": testResult.name,
				"timeStamp": testResult.timeStamp,
				"enabledState": str(testResult.enabledState),
				"processingState": str(testResult.processingState),
				"duration": testResult.duration,
				"logBuffer": testResult.logBuffer.getDataAsPrettyJSON(),
				"description": testResult.description,
			})

		tempFilePath = testResultCollection.createSVG()
		with open(tempFilePath, "r") as f:
			svgLines = f.readlines()
		os.unlink(tempFilePath)

		while (len(svgLines) > 0) and not svgLines[0].startswith("<svg"):
			del svgLines[0]

		content = self.__templateOverview.render(
			svg="".join(svgLines),
			testRecords=jsonTestRecords,
			summary={
				"countTestsPerformed": testResultCollection.countTestsSucceeded + testResultCollection.countTestsFailed,
				"countTestsNotYetPerformed": testResultCollection.countTestsNotYetPerformed,
				"countTestsSucceeded": testResultCollection.countTestsSucceeded,
				"countTestsFailed": testResultCollection.countTestsFailed,
				"totalTestRuntime": testResultCollection.totalTestRuntime,
				"totalTestDuration": testResultCollection.totalTestDuration,
			}
			)
		filePath = os.path.join(outDirPath, "index.html")
		print("Writing: " + filePath)
		with open(filePath, "w") as f:
			f.write(content)
	#

#











from jk_logging import BufferLogger, EnumLogLevel

from .TestCaseInstance import TestCaseInstance, EnumEnabledState, EnumProcessingState





class TestResult(object):

	def __init__(self, testCaseInstance:TestCaseInstance, logBuffer:BufferLogger):
		self.name = testCaseInstance.name
		self.timeStamp = testCaseInstance.timeStamp
		self.enabledState = testCaseInstance.enabledState
		self.processingState = testCaseInstance.processingState
		self.duration = testCaseInstance.duration
		self.logBuffer = logBuffer
		self.description = testCaseInstance.description
	#

#
















#!/usr/bin/env python3
# -*- coding: utf-8 -*-



import time
import collections

import jk_logging

from .Annotations import *
from .TestCaseInstance import *






class TestCaseCollection(object):

	def __init__(self):
		self._allTestCasesByName = {}					# stores names of test cases => test case records
		self._allTestCasesByProvidedVariables = {}		# stores variable names provided by test cases => test case records
		self._allTestCasesAdded = []					# stores test case records
		self._orderOfProcessing = None					# stores all test case records in the order the test cases need to be processed
	#

	#
	# Is the collection prepaired for processing? "Prepaired" means: The dependencies are analysed and the processing order is calculated.
	#
	@property
	def isPrepared(self):
		return self._orderOfProcessing != None
	#

	#
	# Use this test case to the list of test cases.
	#
	def add(self, testCaseInstance):
		assert isinstance(testCaseInstance, TestCaseInstance)

		self._allTestCasesByName[testCaseInstance.name] = testCaseInstance
		self._allTestCasesAdded.append(testCaseInstance)

		for a in testCaseInstance.testCaseAspects:
			if isinstance(a, providesVariable):
				if a.varName in self._allTestCasesByProvidedVariables:
					raise Exception("Test case '" + testCaseInstance.name + "' provides variable '" + a.varName + "' while test case '"
						+ self._allTestCasesByProvidedVariables[a.varName].name + "' already provides variable '" + a.varName + "'!")
				self._allTestCasesByProvidedVariables[a.varName] = testCaseInstance

		# now we are no longer prepared.
		self._unprepare()
	#

	#
	# Retreive the (ordered) list of test tasks to process.
	#
	# @return	TestCaaseInstance[] testCaseInstances		Returns the list of test tasks to process. If the collection is not prepaired, <c>None</c> is returned.
	#
	def getTestTasksToProcess(self):
		if self._orderOfProcessing is None:
			return None
		return list(self._orderOfProcessing)
	#

	def _unprepare(self):
		for testCaseInstance in self._allTestCasesAdded:
			testCaseInstance.reset()
		self._orderOfProcessing = None
	#

	#
	#
	#
	def __buildTestCaseGraph(self, log):
		# 1) set dependencies:
		#		A.runBefore(B)			->  B.dependencies.add(A)
		#		A.runAfter(B)			->  A.dependencies.add(B)
		#		A.requires(B)			->  A.requires.add(B)
		#		A.requiresVariable(xyz)	->  B.dependencies.add(A) where B provides xyz
		for testCaseInstance in self._allTestCasesAdded:
			for a in testCaseInstance.testCaseAspects:

				if isinstance(a, runBefore):
					b = self._allTestCasesByName.get(a.testName, None)
					if b is None:
						log.error("Test case '" + testCaseInstance.name + "' refers to a non-existing test case named '" + a.testName + "'!")
						return False
					testCaseInstance.dependencies.add(a.name)

				elif isinstance(a, runAfter):
					b = self._allTestCasesByName.get(a.testName, None)
					if b is None:
						log.error("Test case '" + testCaseInstance.name + "' refers to a non-existing test case named '" + a.testName + "'!")
						return False
					b.dependencies.add(testCaseInstance.name)

				elif isinstance(a, requires):
					b = self._allTestCasesByName.get(a.testName, None)
					if b is None:
						log.error("Test case '" + testCaseInstance.name + "' refers to a non-existing test case named '" + a.testName + "'!")
						return False
					b.dependencies.add(testCaseInstance.name)
					b.requires.add(testCaseInstance.name)
					testCaseInstance.requires.add(b.name)

				elif isinstance(a, requiresVariable):
					b = self._allTestCasesByProvidedVariables.get(a.varName, None)
					if b is None:
						log.error("Test case '" + testCaseInstance.name + "' refers to variable named '" + a.varName + "' which no test case provides!")
						return False
					testCaseInstance.dependencies.add(b.name)
					testCaseInstance.requires.add(b.name)
	#

	#
	# This method precompiles all test cases in order to execute them later as required.
	#
	def prepare(self, log):
		self.__buildTestCaseGraph(log)

		# 2) check for cycles and determine order of test case execution
		notYetProcessed = list(self._allTestCasesAdded)
		alreadyProcessedNames = set()
		orderOfProcessing = []
		while notYetProcessed:
			notYetProcessedRemaining = []
			for x in notYetProcessed:
				testCaseInstance, bEnabled = x
				if testCaseInstance.areAllDependenciesMet(alreadyProcessedNames):
					alreadyProcessedNames.add(testCaseInstance.name)
					orderOfProcessing.append(x)
				else:
					notYetProcessedRemaining.append(x)
			if len(notYetProcessedRemaining) == len(notYetProcessed):
				# all items have unmet dependencies!
				log.error("Cyclic dependencies detected! The following test cases are involved in this cycle:")
				for x in notYetProcessed:
					testCaseInstance, bEnabled = x
					log.error("Test case '" + testCaseInstance.name + "' depends on: " + repr(testCaseInstance.dependencies))
				return False
			notYetProcessed = notYetProcessedRemaining

		# 3) check if all test cases are enabled that produce output required by other test cases.
		testCaseList = list(orderOfProcessing)
		testCaseList.reverse()						# the last test case will be put first. we process the list in reverse order.
		for x in testCaseList:
			testCaseInstance, bEnabled = x
			if bEnabled:
				print("Test case instance: " + testCaseInstance.name)
				for referringTestCaseName in testCaseInstance.requires:
					print("\tRequires: " + referringTestCaseName)
					self.__enableTestCaseAndPredecessors(referringTestCaseName, testCaseInstance.name, log)

		# we end up here if everything went well.
		self._orderOfProcessing = orderOfProcessing
		return True
	#

	def __enableTestCaseAndPredecessors(self, testCaseName, requiredByTestCaseName, log):
		x = self._allTestCasesByName[testCaseName]
		testCaseInstance, bEnabled = x
		if not bEnabled:
			x[1] = True
			log.notice("Enabling test case '" + testCaseInstance.name + "' as test case '" + requiredByTestCaseName + "' (and maybe other test cases) depends on it.")

		print("Test case instance: " + testCaseInstance.name)
		for referringTestCaseName in testCaseInstance.requires:
			print("\tRequires: " + referringTestCaseName)
			self.__enableTestCaseAndPredecessors(referringTestCaseName, testCaseName, log)
	#

#









